@@ -1,5 +1,5 @@
 import base64
-
+import binascii
 from datetime import timedelta
 
 from django.contrib.auth import authenticate
@@ -7,62 +7,134 @@ from django.utils import timezone
 
 from oauthlib.oauth2 import RequestValidator
 
-from oauth_api.models import get_application_model, AccessToken, AuthorizationCode, RefreshToken
+from oauth_api.models import get_application_model, AccessToken, AuthorizationCode, RefreshToken, AbstractApplication
 from oauth_api.settings import oauth_api_settings
 
-
-Application = get_application_model()
-
-
 GRANT_TYPE_MAPPING = {
-    'authorization_code': (Application.GRANT_AUTHORIZATION_CODE,),
-    'password': (Application.GRANT_PASSWORD,),
-    'client_credentials': (Application.GRANT_CLIENT_CREDENTIALS,),
-    'refresh_token': (Application.GRANT_AUTHORIZATION_CODE, Application.GRANT_PASSWORD,
-                      Application.GRANT_CLIENT_CREDENTIALS)
+    'authorization_code': (AbstractApplication.GRANT_AUTHORIZATION_CODE,),
+    'password': (AbstractApplication.GRANT_PASSWORD,),
+    'client_credentials': (AbstractApplication.GRANT_CLIENT_CREDENTIALS,),
+    'refresh_token': (AbstractApplication.GRANT_AUTHORIZATION_CODE, AbstractApplication.GRANT_PASSWORD,
+                      AbstractApplication.GRANT_CLIENT_CREDENTIALS)
 }
 
+
 class OAuthValidator(RequestValidator):
+    def _get_application(self, client_id, request):
+        """
+        Load application instance for given client_id and store it in request as 'client' attribute
+        """
+        assert hasattr(request, 'client'), "'client' attribute missing from 'request'"
+
+        Application = get_application_model()
+
+        try:
+            request.client = request.client or Application.objects.get(client_id=client_id)
+            return request.client
+        except Application.DoesNotExist:
+            return None
+
+    def _get_auth_string(self, request):
+        auth = request.headers.get('HTTP_AUTHORIZATION', None)
+
+        if not auth:
+            return None
+
+        splitted = auth.split(' ', 1)
+        if len(splitted) != 2:
+            return None
+
+        auth_type, auth_string = splitted
+        if auth_type != 'Basic':
+            return None
+
+        return auth_string
 
     def _authenticate_client_basic(self, request):
         """
         Try authenticating the client using HTTP Basic Authentication method
         """
-        auth = request.headers.get('HTTP_AUTHORIZATION', None)
-
-        if not auth:
+        auth_string = self._get_auth_string(request)
+        if not auth_string:
             return False
 
-        auth_type, auth_string = auth.split(' ')
-        if not auth_type == 'Basic':
-            return False
-
-        encoding = request.encoding or 'utf-8'
-
-        auth_string_decoded = base64.b64decode(auth_string).decode(encoding)
-        client_id, client_secret = auth_string_decoded.split(':', 1)
-
         try:
-            request.client = Application.objects.get(client_id=client_id, client_secret=client_secret)
-            return True
-        except Application.DoesNotExist:
+            encoding = request.encoding or 'utf-8'
+        except AttributeError:
+            encoding = 'utf-8'
+
+        try:
+            b64_decoded = base64.b64decode(auth_string)
+        except (TypeError, binascii.Error):
             return False
 
+        try:
+            auth_string_decoded = b64_decoded.decode(encoding)
+        except UnicodeDecodeError:
+            return False
+
+        client_id, client_secret = auth_string_decoded.split(':', 1)
+
+        if self._get_application(client_id, request) is None:
+            return False
+        elif request.client.client_secret != client_secret:
+            return False
+        else:
+            return True
+
     def _authenticate_client_body(self, request):
         """
         Try authenticating the client using values from request body
         """
-        client_id = request.client_id
-        client_secret = request.client_secret
+        try:
+            client_id = request.client_id
+            client_secret = request.client_secret
+        except AttributeError:
+            return False
 
         if not client_id:
             return False
 
-        try:
-            request.client = Application.objects.get(client_id=client_id, client_secret=client_secret)
-            return True
-        except Application.DoesNotExist:
+        if self._get_application(client_id, request) is None:
             return False
+        elif request.client.client_secret != client_secret:
+            return False
+        else:
+            return True
+
+    def client_authentication_required(self, request, *args, **kwargs):
+        """
+        Determine if client authentication is required for current request.
+
+        According to the rfc6749, client authentication is required in the following cases:
+            - Resource Owner Password Credentials Grant, when Client type is Confidential or when
+              Client was issued client credentials or whenever Client provided client
+              authentication, see `Section 4.3.2`_.
+            - Authorization Code Grant, when Client type is Confidential or when Client was issued
+              client credentials or whenever Client provided client authentication,
+              see `Section 4.1.3`_.
+            - Refresh Token Grant, when Client type is Confidential or when Client was issued
+              client credentials or whenever Client provided client authentication, see
+              `Section 6`_
+
+        :param request: oauthlib.common.Request
+        :return: True or False
+        """
+        if self._get_auth_string(request):
+            return True
+
+        try:
+            if request.client_id and request.client_secret:
+                return True
+        except AttributeError:
+            # Client id or secret not provided
+            pass
+
+        self._get_application(request.client_id, request)
+        if request.client:
+            return request.client.client_type == AbstractApplication.CLIENT_CONFIDENTIAL
+
+        return super(OAuthValidator, self).client_authentication_required(request, *args, **kwargs)
 
     def authenticate_client(self, request, *args, **kwargs):
         """
@@ -80,13 +152,9 @@ class OAuthValidator(RequestValidator):
         Ensure client_id belong to a non-confidential client.
         A non-confidential client is one that is not required to authenticate through other means, such as using HTTP Basic.
         """
-        client_secret = request.client_secret
-        try:
-            request.client = request.client or Application.objects.get(client_id=client_id,
-                                                                       client_secret=client_secret)
-            return request.client.client_type != Application.CLIENT_CONFIDENTIAL
-        except Application.DoesNotExist:
-            return False
+        if self._get_application(client_id, request) is not None:
+            return request.client.client_type != AbstractApplication.CLIENT_CONFIDENTIAL
+        return False
 
     def confirm_redirect_uri(self, client_id, code, redirect_uri, client, *args, **kwargs):
         """
@@ -125,51 +193,83 @@ class OAuthValidator(RequestValidator):
         Persist the authorization_code.
         """
         expires = timezone.now() + timedelta(seconds=oauth_api_settings.ACCESS_TOKEN_EXPIRATION)
-        auth_code = AuthorizationCode(application=request.client, user=request.user, code=code['code'],
-                      expires=expires, redirect_uri=request.redirect_uri,
-                      scope=' '.join(request.scopes))
-        auth_code.save()
+        AuthorizationCode.objects.create(
+            application=request.client,
+            user=request.user,
+            code=code['code'],
+            expires=expires,
+            redirect_uri=request.redirect_uri,
+            scope=' '.join(request.scopes)
+        )
         return request.redirect_uri
 
     def save_bearer_token(self, token, request, *args, **kwargs):
         """
         Persist the Bearer token.
         """
-        if request.refresh_token_object:
-            # Remove used refresh token
+        if request.refresh_token:
+            # Revoke Refresh Token (and related Access Token)
             try:
-                RefreshToken.objects.get(token=request.refresh_token).delete()
+                RefreshToken.objects.get(token=request.refresh_token).revoke()
             except RefreshToken.DoesNotExist:
-                # Already deleted?
-                assert()
+                # Already revoked?
+                pass
 
         expires = timezone.now() + timedelta(seconds=oauth_api_settings.ACCESS_TOKEN_EXPIRATION)
+        user = request.user
         if request.grant_type == 'client_credentials':
-            request.user = request.client.user
+            user = None
 
-        access_token = AccessToken(
-            user=request.user,
+        access_token = AccessToken.objects.create(
+            user=user,
             scope=token['scope'],
             expires=expires,
             token=token['access_token'],
-            application=request.client)
-        access_token.save()
+            application=request.client
+        )
 
         if 'refresh_token' in token:
             if oauth_api_settings.REFRESH_TOKEN_EXPIRATION is not None:
                 expires = timezone.now() + timedelta(seconds=oauth_api_settings.REFRESH_TOKEN_EXPIRATION)
             else:
                 expires = None
-            refresh_token = RefreshToken(
+            RefreshToken.objects.create(
                 user=request.user,
                 token=token['refresh_token'],
                 expires=expires,
                 application=request.client,
-                access_token=access_token)
-            refresh_token.save()
+                access_token=access_token
+            )
 
         return request.client.default_redirect_uri
 
+    def revoke_token(self, token, token_type_hint, request, *args, **kwargs):
+        """
+        Revoke an access or refresh token.
+
+        :param token: The token string.
+        :param token_type_hint: access_token or refresh_token.
+        :param request: The HTTP Request (oauthlib.common.Request)
+        """
+        if token_type_hint not in ['access_token', 'refresh_token']:
+            token_type_hint = None
+
+        token_types = {
+            'access_token': AccessToken,
+            'refresh_token': RefreshToken,
+        }
+
+        token_type = token_types.get(token_type_hint, AccessToken)
+
+        try:
+            token_type.objects.get(token=token, application=request.client).revoke()
+        except token_type.DoesNotExist:
+            # Lookup from all token types except from already looked up type
+            other_types = (_type for _type in token_types.values() if _type != token_type)
+            for other_type in other_types:
+                for token in other_type.objects.filter(token=token, application=request.client):
+                    token.revoke()
+
     def validate_bearer_token(self, token, scopes, request):
         """
         Ensure the Bearer token is valid and authorized access to scopes.
@@ -184,7 +284,7 @@ class OAuthValidator(RequestValidator):
                 request.user = access_token.user
                 request.scopes = scopes
 
-                # Required when autenticating using OAuth2Authentication
+                # Required when authenticating using OAuth2Authentication
                 request.access_token = access_token
                 return True
             return False
@@ -195,11 +295,7 @@ class OAuthValidator(RequestValidator):
         """
         Check that and Application exists with given client_id.
         """
-        try:
-            request.client = request.client or Application.objects.get(client_id=client_id)
-            return True
-        except Application.DoesNotExist:
-            return False
+        return self._get_application(client_id, request) is not None
 
     def validate_code(self, client_id, code, client, request, *args, **kwargs):
         """
@@ -219,7 +315,7 @@ class OAuthValidator(RequestValidator):
         """
         Ensure client is authorized to use the grant_type requested.
         """
-        assert(grant_type in GRANT_TYPE_MAPPING)
+        assert (grant_type in GRANT_TYPE_MAPPING)
         return request.client.authorization_grant_type in GRANT_TYPE_MAPPING[grant_type]
 
     def validate_redirect_uri(self, client_id, redirect_uri, request, *args, **kwargs):
@@ -236,6 +332,7 @@ class OAuthValidator(RequestValidator):
             rt = RefreshToken.objects.select_related('user').get(token=refresh_token)
             if not rt.is_expired:
                 request.user = rt.user
+                request.refresh_token = rt.token
                 request.refresh_token_object = rt
                 return rt.application == client
             return False
@@ -249,9 +346,9 @@ class OAuthValidator(RequestValidator):
         See http://tools.ietf.org/html/rfc6749#section-8.4
         """
         if response_type == 'code':
-            return client.authorization_grant_type == Application.GRANT_AUTHORIZATION_CODE
+            return client.authorization_grant_type == AbstractApplication.GRANT_AUTHORIZATION_CODE
         elif response_type == 'token':
-            return client.authorization_grant_type == Application.GRANT_IMPLICIT
+            return client.authorization_grant_type == AbstractApplication.GRANT_IMPLICIT
         else:
             return False
 
